from __future__ import annotations

import re
from collections import Counter, defaultdict
from dataclasses import dataclass
from pathlib import Path
from typing import Sequence

import spacy
from dateutil.parser import parse, parser
from gp.entity_linking.candidate_recognition._interface import ICanReg
from kgdata.models import Ontology
from sm.dataset import Example, FullTable
from sm.misc.ray_helper import ray_map


class HeuristicCanReg(ICanReg):
    VERSION = 102

    def __call__(
        self, examples: Sequence[Example[FullTable]], ontology: Ontology
    ) -> list[list[int]]:
        return ray_map(
            predict_entity_column,
            [(ex,) for ex in examples],
            verbose=True,
            desc="Predicting entity columns",
            is_func_remote=False,
            using_ray=len(examples) > 1,
        )


num_pattern = r"(\d+((\.)\d+)?)"
unit_pattern = r"(m|in|ft|cm|khz|hz|kw)"

insensitive_patterns = [
    r"^[+-T]?\s*\d+$",  # T3, -1, +1, etc.
    r"^[\[\(]\s*\d+\s*[\]\)]$",  # (1), [1], etc.
    r"^\d+(\s*[-=,]\s*\d+)+$",  # 1=2 | 1,2 | 1-2
    r"^\d+\s*\(\s*\d+\s*\)$",  # 1 (1)
    r"^(-|•|x|n/a)$",  # missing values
    # r".*\(\s*\d+(\s*,\s*\d+)+\s*\)?",  # (1, 2, 3)
    rf"{num_pattern}\s*{unit_pattern}",  # 1.2 m, 1.2 in, etc.
    rf"{num_pattern}\s*{unit_pattern}\s*{num_pattern}\s*{unit_pattern}",  # 5 ft 11 in
    rf"{num_pattern}\s*{unit_pattern}\s*\(\s*{num_pattern}\s*{unit_pattern}\s*{num_pattern}\s*{unit_pattern}\s*\)",  # 5 ft 11 in
]
regexes = [re.compile(p, flags=re.IGNORECASE) for p in insensitive_patterns]


def get_type(cell: str | int | float):
    global regexes

    if isinstance(cell, (int, float)):
        return DataType.NUM

    normcell = cell.strip()
    normcell = normcell.replace("−", "-").replace("–", "-")

    dtype = MSpacy.get_instance().get_type(normcell)
    if dtype != DataType.TEXT:
        return dtype
    if any(reg.match(normcell) is not None for reg in regexes):
        return DataType.NUM
    return dtype


def predict_entity_column(ex: Example[FullTable]) -> list[int]:
    ent_cols = []
    for col in ex.table.table.columns:
        dtype_counter = defaultdict(int)
        for ri, cell in enumerate(col.values):
            dtype_cell = get_type(cell)
            dtype_counter[dtype_cell] += 1

        temp_dtype = sorted(dtype_counter.items(), key=lambda x: x[1], reverse=True)
        if not len(temp_dtype):
            # number
            continue
        else:
            if (
                temp_dtype[0][0] == DataType.NONE
                and len(temp_dtype) > 1
                and temp_dtype[1][1]
            ):
                col_dtype = temp_dtype[1][0]
            else:
                col_dtype = temp_dtype[0][0]

        if col_dtype == DataType.TEXT:
            ent_cols.append(col.index)
    return ent_cols


class DataType:
    TEXT = 1
    NUM = 0
    NONE = 2


class MSpacy(object):
    """
    Code borrow from Phuc Nguyen. mtab.api.annotator.m_spacy
    """

    instance = None

    def __init__(self):
        self.model_names = {
            "en": "en_core_web_sm",  # sm, md, lg, trf
            "all": "xx_sent_ud_sm",  # python -m spacy download xx_sent_ud_sm
        }
        self.models = {
            "en": None,
            "all": None,
        }
        self.load_model(lang="en")
        self.load_model(lang="all")
        self.dims_num = {
            "CARDINAL",
            "PERCENT",
            "MONEY",
            "ORDINAL",
            "QUANTITY",
            "TIME",
            "DATE",
        }
        self.NONE_CELLS = {
            "''",
            '""',
            "-",
            "--",
            "'-'",
            '"-"',
            " ",
            ".",
            "' '",
            '" "',
            "nan",
            "none",
            "null",
            "blank",
            "yes",
            "unknown",
            "?",
            "??",
            "???",
            "0",
            "total",
        }
        self.NUM_SPACY = {
            "CARDINAL",
            "PERCENT",
            "MONEY",
            "ORDINAL",
            "QUANTITY",
            "TIME",
            "DATE",
        }

    @staticmethod
    def get_instance():
        if MSpacy.instance is None:
            MSpacy.instance = MSpacy()
        return MSpacy.instance

    def load_model(self, lang):
        """Load SpaCy model with the language of [lang]

        Args:
            lang (string): language id
        """
        if lang != "en":
            lang = "all"
        print("Loaded: Type entity model (%s)" % self.model_names[lang])
        self.models[lang] = spacy.load(self.model_names[lang])  # type: ignore

    def _parse(self, text, lang="en"):
        """[summary]

        Args:
            text ([type]): [description]
            lang (str, optional): [description]. Defaults to "en".

        Returns:
            [type]: [description]
        """

        spacy_model = self.models.get(lang, None)
        if not spacy_model:
            self.load_model(lang)
            spacy_model = self.models[lang]
        response = spacy_model(text)  # type: ignore
        return response

    def get_majority_type(self, text, lang="en"):
        """[summary]

        Args:
            text ([type]): [description]
            lang (str, optional): [description]. Defaults to "en".

        Returns:
            [type]: [description]
        """
        types = DataType.TEXT
        text = str(text)
        if text:
            # count_numbers = sum(c.isdigit() for c in text)
            # if count_numbers > len(text):
            response = self._parse(text, lang)
            if len(response.ents):
                num_len = 0
                num_types = Counter()
                text_types = Counter()
                for r in response.ents:
                    if r.label_ in self.dims_num:
                        num_len += len(r.text)
                        num_types[r.label_] += 1
                    else:
                        text_types[r.label_] += 1
                if num_len > 0.5 * len(text):
                    types = num_types.most_common(1)[0][0]
                elif len(text_types) > 0:
                    types = text_types.most_common(1)[0][0]
        return types

    def get_type(self, text, lang="en"):
        if not text or text.lower() in self.NONE_CELLS:
            return DataType.NONE

        if text[0] == "〒" or text[:2] == "代表":
            return DataType.NUM

        txt_tmp = text.replace("%", "").replace(" ", "")
        if self.convert_num(txt_tmp) is not None:
            return DataType.NUM

        if self.is_date(text):
            return DataType.NUM

        main_type = self.get_majority_type(text, lang)

        if main_type in self.NUM_SPACY:
            return DataType.NUM

        return DataType.TEXT

    def is_number(self, text, lang="en"):
        is_num = False
        if len(text):
            response = self._parse(text, lang)
            if len(response.ents):
                num_len = 0
                for r in response.ents:
                    if r.label_ in self.dims_num:
                        num_len += len(r.text)
                if num_len > 0.5 * len(text):
                    is_num = True
        return is_num

    def is_text(self, text, lang="en"):
        return not self.is_number(text, lang)

    def get_entity_text(self, text, lang="en"):
        results = []
        if len(text):
            response = self._parse(text, lang)
            if len(response.ents):
                for r in response.ents:
                    if r.label_ not in self.dims_num:
                        results.append((r.text, r.label_))
        return results

    @classmethod
    def convert_num(cls, text):
        if not text:
            return None
        try:
            text = cls.removeCommasBetweenDigits(text)
            # tmp = representsFloat(text)
            # if not tmp:
            #     return None
            #
            # return parseNumber(text)
            return float(text)  # type: ignore
        except ValueError:
            return None

    @classmethod
    def removeCommasBetweenDigits(cls, text):
        """
        :example:
        >>> removeCommasBetweenDigits("sfeyv dsf,54dsf ef 6, 6 zdgy 6,919 Photos and 3,3 videos6,")
        'sfeyv dsf,54dsf ef 6, 6 zdgy 6919 Photos and 33 videos6,'
        """
        if text is None:
            return None
        else:
            return re.sub(r"([0-9]),([0-9])", "\g<1>\g<2>", text)

    @classmethod
    def is_date_complete(cls, date, must_have_attr=("year", "month")):
        parse_res, _ = parser()._parse(date)  # type: ignore
        # If date is invalid `_parse` returns (None,None)
        if parse_res is None:
            return False
        # For a valid date `_result` object is returned. E.g. _parse("Sep 23") returns (_result(month=9, day=23), None)
        for attr in must_have_attr:
            if getattr(parse_res, attr) is None:
                return False
        return True

    @classmethod
    def is_date(cls, string, fuzzy=False):
        if not string:
            return False
        try:
            parse(string, fuzzy=fuzzy)
            if cls.is_date_complete(string):
                return True
            return False
        except Exception:
            return False



from gp.actors.el.cangen import CanGenActor, CanGenActorArgs
from gp.actors.el.canrank import CanRankActor, CanRankActorArgs
from gp.actors.el.canreg import CanRegActor, CanRegActorArgs

__all__ = [
    "CanGenActor",
    "CanGenActorArgs",
    "CanRankActor",
    "CanRankActorArgs",
    "CanRegActor",
    "CanRegActorArgs",
]



from __future__ import annotations

from dataclasses import dataclass
from typing import Callable, Mapping, Optional, TypeAlias

from gp_core.algorithms import CanGraphExtractedResult
from graph.retworkx import BaseEdge, BaseNode, RetworkXStrDiGraph
from sm.inputs.table import ColumnBasedTable
from sm.misc.funcs import assert_not_null
from sm.namespaces.namespace import KnowledgeGraphNamespace
from sm.outputs.semantic_model import LiteralNodeDataType
from sm.prelude import O


@dataclass
class CGNode(BaseNode[str]):
    id: str  # it is number but converted into a string

    is_statement_node: bool
    is_column_node: bool
    is_entity_node: bool
    is_literal_node: bool
    is_in_context: bool

    column_index: Optional[int]  # not null if is_column_node=True

    def to_sm_node(
        self,
        table: ColumnBasedTable,
        kgns: KnowledgeGraphNamespace,
        id2literal: Callable[[str], str],
    ):
        """Convert a node in the candidate graph to a node in the semantic model"""
        if self.is_statement_node:
            return O.ClassNode(
                abs_uri=kgns.statement_uri,
                rel_uri=kgns.get_rel_uri(kgns.statement_uri),
                approximation=False,
                readable_label="Statement",
            )
        if self.is_column_node:
            assert self.column_index is not None
            return O.DataNode(
                col_index=self.column_index,
                label=assert_not_null(
                    table.get_column_by_index(self.column_index).clean_multiline_name
                ),
            )
        if self.is_entity_node:
            entid = id2literal(self.id)
            return O.LiteralNode(
                value=kgns.uri_to_id(entid),
                datatype=LiteralNodeDataType.Entity,
            )
        if self.is_literal_node:
            litval = id2literal(self.id)
            return O.LiteralNode(
                value=litval,
                datatype=LiteralNodeDataType.String,
            )
        raise RuntimeError("Unknown node type")


CGEdgeTriple: TypeAlias = tuple[str, str, str]
CGEdge: TypeAlias = BaseEdge[str, str]


class CGraph(RetworkXStrDiGraph[str, CGNode, BaseEdge[str, str]]):
    """A temporary copied version of the candidate graph in Rust for running post-processing"""

    @staticmethod
    def from_rust(cangraph: CanGraphExtractedResult) -> CGraph:
        g = CGraph()
        for u in cangraph.nodes:
            uid = u.id()
            is_in_context = False
            is_column = u.is_column()
            is_entity = u.is_entity()
            is_literal = u.is_literal()
            column_index = None
            if is_entity:
                is_in_context = assert_not_null(u.try_as_entity()).is_in_context
            if is_literal:
                is_in_context = assert_not_null(u.try_as_literal()).is_in_context
            if is_column:
                column_index = assert_not_null(u.try_as_column()).column

            g.add_node(
                CGNode(
                    str(uid),
                    u.is_statement(),
                    is_column,
                    is_entity,
                    is_literal,
                    is_in_context,
                    column_index,
                )
            )
        for row in cangraph.edgedf.iter_rows(named=True):
            g.add_edge(
                BaseEdge(
                    -1,
                    str(row["source"]),
                    str(row["statement"]),
                    cangraph.edges[row["inedge"]],
                )
            )
            g.add_edge(
                BaseEdge(
                    -1,
                    str(row["statement"]),
                    str(row["target"]),
                    cangraph.edges[row["outedge"]],
                )
            )
        return g

    def remove_dangling_statement(self):
        """Remove statement nodes that have no incoming nodes or no outgoing edges"""
        for s in self.nodes():
            if s.is_statement_node and (
                self.in_degree(s.id) == 0 or self.out_degree(s.id) == 0
            ):
                self.remove_node(s.id)

    def remove_standalone_nodes(self):
        """Remove nodes that do not any incoming / outgoing edges"""
        for u in self.nodes():
            if self.degree(u.id) == 0:
                self.remove_node(u.id)



from __future__ import annotations

from collections import defaultdict
from operator import itemgetter
from typing import Sequence

import sm.misc as M
from graph.retworkx.api import digraph_all_simple_paths
from sm.dataset import FullTable
from sm.misc.funcs import assert_not_null

from gramsplus.semanticmodeling.postprocessing.cgraph import CGraph
from gramsplus.semanticmodeling.postprocessing.interface import (
    EdgeProb,
    NodeProb,
    PostprocessingFn,
    PostprocessingResult,
)


class GreedyKnownTargetsFn(PostprocessingFn):
    """Post-proessing the select the binary relationships and types between given columns"""

    def __init__(
        self,
        cg: CGraph,
        edge_targets: Sequence[tuple[int, int]],
        type_targets: Sequence[int],
        cpa_threshold: float,
        cta_threshold: float,
    ):
        self.cg = cg
        self.edge_targets = edge_targets
        self.type_targets = set(type_targets)
        self.cpa_threshold = cpa_threshold
        self.cta_threshold = cta_threshold

    def __call__(
        self,
        node_probs: NodeProb,
        edge_probs: EdgeProb,
    ) -> PostprocessingResult:
        nodes = {}
        for colindex, utypes in node_probs.items():
            if colindex in self.type_targets:
                ctype, cscore = max(utypes.items(), key=itemgetter(1))
                if cscore >= self.cta_threshold:
                    nodes[colindex] = ctype

        column2node = {
            assert_not_null(u.column_index): u.id
            for u in self.cg.iter_nodes()
            if u.is_column_node
        }
        selected_edges = []

        for sci, tci in self.edge_targets:
            uid = column2node[sci]
            vid = column2node[tci]

            if not self.cg.has_node(uid) or not self.cg.has_node(vid):
                continue
            paths = digraph_all_simple_paths(self.cg, uid, vid, cutoff=2)
            paths = [
                (inedge, outedge)
                for inedge, outedge in paths
                if inedge.key == outedge.key
                and edge_probs[inedge.source, inedge.target, inedge.key]
                >= self.cpa_threshold
                and edge_probs[outedge.source, outedge.target, outedge.key]
                >= self.cta_threshold
            ]
            if len(paths) == 0:
                continue

            best_path = max(
                paths,
                key=lambda p: sum(edge_probs[e.source, e.target, e.key] for e in p),
            )
            for edge in best_path:
                selected_edges.append((edge.source, edge.target, edge.key))

        return PostprocessingResult(selected_edges, nodes)



import enum
from typing import Optional, TypeAlias

from kgdata_core.models import EntityMetadata, Property, StatementView, Value, ValueView

from gp_core.literal_matchers import ParsedTextRepr

class LocalGramsDB:
    def __init__(self, datadir: str) -> None: ...
    @staticmethod
    def init(datadir: str) -> None: ...
    @staticmethod
    def get_instance() -> LocalGramsDB: ...
    def get_algo_context(
        self, table: Table, n_hop: int, parallel: bool
    ) -> AlgoContext: ...
    def get_property(self, id: str) -> Property: ...
    def has_entity(self, id: str) -> bool: ...
    def get_redirected_entity_id(self, id: str) -> Optional[str]: ...
    def get_entity_metadata(self, id: str) -> EntityMetadata: ...
    def get_entity_pagerank(self, id: str) -> float: ...

class RemoteGramsDB:
    def __init__(
        self,
        datadir: str,
        entity_urls: list[str],
        entity_metadata_urls: list[str],
        entity_batch_size: int,
        entity_metadata_batch_size: int,
    ) -> None: ...
    @staticmethod
    def init(
        datadir: str,
        entity_urls: list[str],
        entity_metadata_urls: list[str],
        entity_batch_size: int,
        entity_metadata_batch_size: int,
    ) -> None: ...
    @staticmethod
    def get_instance() -> LocalGramsDB: ...
    def get_algo_context(
        self, table: Table, n_hop: int, parallel: bool
    ) -> AlgoContext: ...
    def get_property(self, id: str) -> Property: ...
    def has_entity(self, id: str) -> bool: ...
    def get_redirected_entity_id(self, id: str) -> Optional[str]: ...
    def get_entity_metadata(self, id: str) -> EntityMetadata: ...
    def get_entity_pagerank(self, id: str) -> float: ...

GramsDB: TypeAlias = LocalGramsDB | RemoteGramsDB

class AlgoContext:
    def get_entity_statement(
        self, entity_id: str, prop: str, stmt_index: int
    ) -> StatementView: ...

class Table:
    id: str
    links: list[list[list[Link]]]
    columns: list[Column]
    context: Context

    def __init__(
        self,
        id: str,
        links: list[list[list[Link]]],
        columns: list[Column],
        context: Context,
    ) -> None: ...
    def get_links(self, row: int, col: int) -> list[Link]: ...
    def save(self, outfile: str) -> None: ...

class TableCells:
    cells: list[list[ParsedTextRepr]]

    def __init__(self, cell: list[list[ParsedTextRepr]]) -> None: ...
    def save(self, outfile: str) -> None: ...

class Link:
    start: int
    end: int
    url: Optional[str]
    entities: list[EntityId]
    candidates: list[CandidateEntityId]

    def __init__(
        self,
        start: int,
        end: int,
        url: Optional[str],
        entities: list[EntityId],
        candidates: list[CandidateEntityId],
    ) -> None: ...

class Column:
    index: int
    name: Optional[str]
    values: list[str]

    def __init__(self, index: int, name: Optional[str], values: list[str]) -> None: ...

class Context:
    page_title: Optional[str]
    page_url: Optional[str]
    page_entities: Optional[list[EntityId]]

    def __init__(
        self,
        page_title: Optional[str],
        page_url: Optional[str],
        page_entities: Optional[list[EntityId]],
    ) -> None: ...

class CandidateEntityId:
    id: EntityId
    probability: float

    def __init__(self, id: EntityId, probability: float) -> None: ...

class EntityId(tuple[str]):
    def __init__(self, id: str) -> None: ...
    @property
    def id(self) -> str: ...

class MatchMethod(enum.Enum):
    LiteralMatching = enum.auto()
    LinkMatching = enum.auto()

class CGNode:
    def __init__(self): ...
    def id(self) -> int: ...
    def is_column(self) -> bool: ...
    def is_statement(self) -> bool: ...
    def is_entity(self) -> bool: ...
    def is_literal(self) -> bool: ...
    def try_as_column(self) -> Optional[CGColumnNode]: ...
    def try_as_statement(self) -> Optional[CGStatementNode]: ...
    def try_as_entity(self) -> Optional[CGEntityNode]: ...
    def try_as_literal(self) -> Optional[CGLiteralNode]: ...

class CGColumnNode:
    id: int
    label: str
    column: int

class CGStatementNode:
    id: int

class CGEntityNode:
    id: int
    entity_id: str
    entity_prob: float
    is_in_context: bool

class CGLiteralNode:
    id: int
    value: ValueView
    value_prob: float
    is_in_context: bool

__all__ = [
    "LocalGramsDB",
    "AlgoContext",
    "Table",
    "TableCells",
    "Link",
    "Column",
    "Context",
    "EntityId",
    "CandidateEntityId",
    "Value",
    "MatchMethod",
    "CGNode",
    "CGColumnNode",
]
]



from gp.actors.data._data import DataActor, DataActorArgs, GPExample, PredictionTargets
from gp.actors.data._db import KGDB, DBActor, DBActorArgs, KGDBArgs, KGName

__all__ = [
    "DataActor",
    "DataActorArgs",
    "GPExample",
    "KGDB",
    "DBActor",
    "DBActorArgs",
    "KGDBArgs",
    "KGName",
    "PredictionTargets",
]



from __future__ import annotations

from abc import abstractmethod
from typing import Sequence

from kgdata.models import Ontology
from sm.dataset import Example, FullTable


class ICanReg:
    @abstractmethod
    def __call__(
        self, examples: Sequence[Example[FullTable]], ontology: Ontology
    ) -> list[list[int]]:
        pass



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Mapping, Sequence

from gp.entity_linking.candidate_recognition._interface import ICanReg
from gp.misc.evaluation.sm_wikidata import SemanticModelHelper
from kgdata.models import Ontology
from kgdata.models.ont_property import OntologyProperty
from sm.dataset import Example, FullTable
from sm.namespaces.namespace import KnowledgeGraphNamespace, OutOfNamespace
from sm.typing import InternalID


class OracleCanReg(ICanReg):
    VERSION = 102

    def __init__(
        self,
        include_should_be_entity_columns: bool,
        include_link_columns: bool,
    ):
        """
        Args:
            include_should_be_entity_columns: Whether to include should_be_entity_columns in the prediction
            include_link_columns: Whether to include columns that contain links in the prediction
        """
        self.include_should_be_entity_columns = include_should_be_entity_columns
        self.include_link_columns = include_link_columns

    def __call__(
        self, examples: Sequence[Example[FullTable]], ontology: Ontology
    ) -> list[list[int]]:
        output = []
        for example in examples:
            entity_columns = []
            for col in example.table.table.columns:
                ci = col.index
                if is_column_entity(example, ci):
                    entity_columns.append(ci)
                elif self.include_should_be_entity_columns and should_be_column_entity(
                    example, ci, ontology.props, ontology.kgns
                ):
                    entity_columns.append(ci)
                elif self.include_link_columns and column_has_links(example, ci):
                    entity_columns.append(ci)
            output.append(entity_columns)
        return output


def is_column_entity(example: Example[FullTable], ci: int):
    for sm in example.sms:
        if not sm.has_data_node(ci):
            continue
        stypes = sm.get_semantic_types_of_column(ci)
        if any(
            stype.predicate_abs_uri in SemanticModelHelper.ID_PROPS for stype in stypes
        ):
            return True
    return False


def column_has_links(example: Example[FullTable], ci: int):
    for celllinks in example.table.links[:, ci]:
        if any(len(link.entities) > 0 for link in celllinks):
            return True
    return False


def should_be_column_entity(
    example: Example[FullTable],
    ci: int,
    props: Mapping[InternalID, OntologyProperty],
    kgns: KnowledgeGraphNamespace,
):
    for sm in example.sms:
        if not sm.has_data_node(ci):
            continue
        dnode = sm.get_data_node(ci)
        for inedge in sm.in_edges(dnode.id):
            try:
                propid = kgns.uri_to_id(inedge.abs_uri)
            except OutOfNamespace:
                continue

            try:
                if props[propid].is_object_property():
                    return True
            except KeyError:
                pass
    return False



from __future__ import annotations

from abc import abstractmethod
from copy import copy
from dataclasses import dataclass
from functools import cached_property
from operator import itemgetter
from pathlib import Path
from typing import Optional

import lightning.pytorch as pl
import torch
from gramsplus.actors.candidate_graph import CanGraphActor
from gramsplus.actors.candidate_ranking import CanRankActor
from gramsplus.misc.dataset import ColumnarDataset
from gramsplus.semanticmodeling.postprocessing.interface import EdgeProb, NodeProb
from ream.actors.base import BaseActor
from ream.cache_helper import Cache, MemBackend
from ream.data_model_helper import NumpyDataModelHelper
from ream.dataset_helper import DatasetList
from ream.helper import import_attr
from scripts.www24.cpa_helper import make_dataset
from scripts.www24.cta_helper import predict_column_prob, predict_cta
from scripts.www24.stores.cpa_store import CPAStore
from scripts.www24.stores.cpa_store_fn import CPAStoreFn, CPAStoreFnArgs
from sm.misc.funcs import import_attr
from torch.utils.data import DataLoader
from tqdm.auto import tqdm


@dataclass
class CPAStoreInvokeArgs:
    args: CPAStoreFnArgs
    funcs: list[str]

    @cached_property
    def enable_fns(self) -> list[type[CPAStoreFn]]:
        return [import_attr(fn) for fn in self.funcs]


@dataclass
class PretrainedCPAModelArgs:
    model_class: str
    model_file: Path
    store_invoke_args: CPAStoreInvokeArgs


class PretrainedCPAModel(BaseActor[PretrainedCPAModelArgs]):
    VERSION = 100

    def __init__(
        self,
        params: PretrainedCPAModelArgs,
        store: CPAStore,
    ):
        super().__init__(params, [store])
        self.store = store

    @Cache.cache(backend=MemBackend())
    def __call__(self, dsquery: str, verbose: bool) -> DatasetList[EdgeProb]:
        args = copy(self.params.store_invoke_args.args)
        args.dsquery = dsquery
        return self.method.predict_dataset(
            self.store,
            CPAStoreInvokeArgs(args, self.params.store_invoke_args.funcs),
            verbose,
        )

    @cached_property
    def method(self):
        self.model_class: type[BaseTorchCPAModel] = import_attr(self.params.model_class)
        if torch.cuda.is_available():
            map_location = "cuda"
        else:
            map_location = "cpu"
        return self.model_class.load_from_checkpoint(
            self.params.model_file, map_location=map_location
        )


class BaseTorchCPAModel(pl.LightningModule):
    EVAL_BATCH_SIZE = 100
    EXPECTED_EVAL_ARGS = set()

    def make_dataset(
        self, store: CPAStore, store_invoke_args: CPAStoreInvokeArgs
    ) -> ColumnarDataset:
        ds = store(store_invoke_args.args, store_invoke_args.enable_fns)
        return make_dataset(ds, is_train=False)

    def predict_dataset(
        self, store: CPAStore, store_invoke_args: CPAStoreInvokeArgs, verbose: bool
    ) -> DatasetList[EdgeProb]:
        self.eval()

        params = next(self.parameters())
        device = params.device

        dsquery = store_invoke_args.args.dsquery

        dataset = self.make_dataset(store, store_invoke_args)
        dloader = DataLoader(
            dataset,
            batch_size=self.EVAL_BATCH_SIZE,
            shuffle=False,
            pin_memory=params.is_cuda,
            collate_fn=dataset.collate_fn,
        )
        with torch.no_grad():
            probs = []
            for batch in tqdm(
                dloader,
                total=len(dloader),
                desc="predicting cpa",
                disable=not verbose,
            ):
                kwargs = {}
                for arg in self.EXPECTED_EVAL_ARGS:
                    kwargs[arg] = batch[arg].to(device)
                output = self.forward(**kwargs)
                probs.append(output.prob.cpu())

            probs = torch.cat(probs).numpy()

        examples = store.data_actor(dsquery)
        cangraphs = store.cangraph_actor(dsquery)
        index = NumpyDataModelHelper.create_simple_index(
            [ex.table.table.table_id for ex in examples],
            [len(c.edgedf) for c in cangraphs],
        )

        output = []
        for ex, cangraph in zip(examples, cangraphs):
            edge_probs = {}
            start, end = index[ex.table.table.table_id]
            for ri, row in enumerate(cangraph.edgedf.iter_rows(named=True)):
                ri = ri + start
                source = str(row["source"])
                target = str(row["target"])
                statement = str(row["statement"])
                inedge = cangraph.edges[row["inedge"]]
                outedge = cangraph.edges[row["outedge"]]

                edge_probs[statement, target, outedge] = probs[ri]
                if outedge == inedge:
                    edge_probs[source, statement, inedge] = probs[ri]
            output.append(edge_probs)
        return DatasetList(examples.name, output)


@dataclass
class PretrainedCTAModelArgs:
    model_class: str
    model_args: dict


class PretrainedCTAModel(BaseActor[PretrainedCTAModelArgs]):
    VERSION = 101

    MOD_SELECTED_TYPE_SCORE = 100

    def __init__(self, params: PretrainedCTAModelArgs, cangraph_actor: CanGraphActor):
        super().__init__(params, [cangraph_actor])
        self.cangraph_actor = cangraph_actor

    @Cache.cache(backend=MemBackend())
    def __call__(self, dsquery: str, verbose: bool) -> DatasetList[NodeProb]:
        return self.method.predict_dataset(
            self.cangraph_actor, dsquery, self.params.model_args, verbose
        )

    @cached_property
    def method(self):
        self.model_class: type[HeuristicELCTAModel] = import_attr(
            self.params.model_class
        )
        return self.model_class()

    def cta_score_offset(self) -> float:
        return self.method.cta_score_offset(self.params.model_args)


class HeuristicELCTAModel:
    def predict_dataset(
        self, actor: CanGraphActor, dsquery: str, args: dict, verbose: bool
    ) -> DatasetList[NodeProb]:
        examples = actor.data_actor(dsquery)
        candidates = actor.canrank_actor.get_candidate_entities(dsquery)

        if args.get("topk", None) is not None:
            candidates = candidates.top_k_candidates(args["topk"])

        ress = predict_cta(
            actor.data_actor.get_kgdb(dsquery),
            examples,
            candidates,
            predict_column_prob,
            {"max_extend_distance": args["max_extend_distance"]},
        )

        output = []
        for res in ress:
            nodeprob = {}
            for ci, ctypes in res.col2type_freq.items():
                nodeprob[ci] = {ctype: cscore for ctype, cscore in ctypes}

                if args["modify_selected_type_score"]:
                    # changing the score of the correct type to be the highest one + 100
                    max_score = max(ctypes, key=itemgetter(1))[1]
                    assert max_score < 100
                    type, score = res.col2type[ci]
                    score = score + 100
                    nodeprob[ci][type] = score
            output.append(nodeprob)
        return DatasetList(examples.name, output)

    def cta_score_offset(self, args: dict) -> float:
        if args["modify_selected_type_score"]:
            return 100
        return 0.0



from __future__ import annotations

from dataclasses import dataclass, field

from gp.entity_linking.candidate_recognition import ICanReg, OracleCanReg
from kgdata.models import Ontology
from libactor.actor import Actor
from libactor.cache import BackendFactory, IdentObj, cache
from sm.dataset import Example, FullTable
from sm.misc.funcs import import_attr


@dataclass
class CanRegActorArgs:
    clspath: str | type
    clsargs: dict | object = field(default_factory=dict)


class CanRegActor(Actor[CanRegActorArgs]):
    VERSION = 103

    def forward(
        self, input: IdentObj[Example[FullTable]], ontology: IdentObj[Ontology]
    ) -> IdentObj[list[int]]:
        ent_cols = self.invoke(input, ontology)
        return IdentObj(str(ent_cols), ent_cols)

    @cache(
        backend=BackendFactory.actor.sqlite.pickle(mem_persist=True),
    )
    def invoke(
        self, example: IdentObj[Example[FullTable]], ontology: IdentObj[Ontology]
    ) -> list[int]:
        return self.get_method()([example.value], ontology.value)[0]

    def is_oracle_entity_linking(self) -> bool:
        return isinstance(self.get_method(), OracleCanReg)

    @cache(backend=BackendFactory.actor.mem)
    def get_method(self) -> ICanReg:
        if isinstance(self.params.clspath, str):
            cls = import_attr(self.params.clspath)
        else:
            cls = self.params.clspath
        if isinstance(self.params.clsargs, dict):
            return cls(**self.params.clsargs)
        if isinstance(self.params.clsargs, tuple):
            return cls(*self.params.clsargs)
        return cls(self.params.clsargs)



from __future__ import annotations

import gp_core as gcore
from sm.dataset import Example, FullTable
from sm.inputs.column import Column
from sm.inputs.link import Link

from gp.el.candidate_generation.cangen_model import TableCandidateEntities


def to_rust_table(
    ex: Example[FullTable], cans: TableCandidateEntities
) -> gcore.models.Table:
    def to_col(col: Column) -> gcore.models.Column:
        values = []
        for v in col.values:
            if isinstance(v, str):
                values.append(v)
            elif v is None:
                values.append("")
            else:
                raise ValueError(f"Unsupported value type: {type(v)}")
        return gcore.models.Column(col.index, col.clean_multiline_name, values)

    def to_links(ri: int, ci: int, links: list[Link]) -> list[gcore.models.Link]:
        if cans.has_cell_candidates(ri, ci):
            cell_cans = cans.get_cell_candidates(ri, ci)
            candidates = [
                gcore.models.CandidateEntityId(
                    gcore.models.EntityId(cell_cans.id[i]),
                    cell_cans.score[i],
                )
                for i in range(len(cell_cans))
            ]
        else:
            cell_cans = None
            candidates = []

        if len(links) == 0:
            if len(candidates) > 0:
                return [
                    gcore.models.Link(
                        start=0,
                        end=len(ex.table.table[ri, ci]),
                        url=None,
                        entities=[],
                        candidates=candidates,
                    )
                ]
            return []

        return [
            gcore.models.Link(
                start=0,
                end=len(ex.table.table[ri, ci]),
                url=None,
                entities=[
                    gcore.models.EntityId(entid)
                    for entid in {entid for link in links for entid in link.entities}
                ],
                candidates=candidates,
            )
        ]

    return gcore.models.Table(
        ex.table.table.table_id,
        [
            [to_links(ri, ci, links) for ci, links in enumerate(row)]
            for ri, row in enumerate(ex.table.links.data)
        ],
        [to_col(col) for col in ex.table.table.columns],
        gcore.models.Context(
            None,
            None,
            [],
        ),
    )
    )



from __future__ import annotations

import re
from dataclasses import dataclass, field
from datetime import MINYEAR, datetime
from typing import Optional

import ftfy
import gp_core.literal_matchers as gcore_literal_matchers
from dateutil.parser import ParserError
from dateutil.parser import parse as dt_parse
from numparser import FSMParser, ParsedNumber
from numparser.fsm.parser import DEFAULT_NOT_UNITS


@dataclass
class TextParserConfigs:
    """Configurations for the parser"""

    NUM_PARSER: str = field(
        default="gramsplus.semanticmodeling.text_parser.BasicNumberParser",
        metadata={"help": "number parser"},
    )
    DATETIME_PARSER: str = field(
        default="gramsplus.semanticmodeling.text_parser.BasicDatetimeParser",
        metadata={"help": "datetime parser"},
    )


@dataclass
class ParsedTextRepr:
    origin: str
    normed_string: str
    number: Optional[ParsedNumber]
    datetime: Optional[ParsedDatetimeRepr]

    def to_rust(self) -> gcore_literal_matchers.ParsedTextRepr:
        return gcore_literal_matchers.ParsedTextRepr(
            self.origin,
            self.normed_string,
            (
                None
                if self.number is None
                else gcore_literal_matchers.ParsedNumberRepr(
                    self.number.number,
                    self.number.number_string,
                    isinstance(self.number.number, int),
                    self.number.unit,
                    self.number.prob,
                )
            ),
            None if self.datetime is None else self.datetime.to_rust(),
        )


@dataclass
class ParsedDatetimeRepr:
    year: Optional[int] = None
    month: Optional[int] = None
    day: Optional[int] = None
    hour: Optional[int] = None
    minute: Optional[int] = None
    second: Optional[int] = None

    def has_only_year(self):
        return (
            self.year is not None
            and self.month is None
            and self.day is None
            and self.hour is None
            and self.minute is None
            and self.second is None
        )

    def first_day_of_year(self):
        return (
            self.year is not None
            and self.month == 1
            and self.day == 1
            and self.hour is None
            and self.minute is None
            and self.second is None
        )

    def to_rust(self) -> gcore_literal_matchers.ParsedDatetimeRepr:
        return gcore_literal_matchers.ParsedDatetimeRepr(
            self.year,
            self.month,
            self.day,
            self.hour,
            self.minute,
            self.second,
        )


class TextParser:
    def __init__(self, cfg: TextParserConfigs):
        self.cache = {}
        self.cfg = cfg

        if (
            cfg.DATETIME_PARSER
            == "gramsplus.semanticmodeling.text_parser.BasicDatetimeParser"
        ):
            self.dt_parser = BasicDatetimeParser()
        else:
            raise NotImplementedError()

        if cfg.NUM_PARSER == "gramsplus.semanticmodeling.text_parser.BasicNumberParser":
            self.number_parser = BasicNumberParser()
        else:
            raise NotImplementedError()

        self.number_chars = re.compile(r"[^0-9\.+-]")

    @staticmethod
    def default():
        return TextParser(TextParserConfigs())

    def parse(self, text: str) -> ParsedTextRepr:
        if text not in self.cache:
            normed_text = self._norm_string(text)
            dt = self.dt_parser.parse(text, normed_text)
            number = self.number_parser.parse(text, normed_text)

            self.cache[text] = ParsedTextRepr(
                origin=text,
                normed_string=normed_text,
                number=number,
                datetime=dt,
            )
        return self.cache[text]

    def _norm_string(self, text: str):
        return ftfy.fix_text(text).replace("\xa0", " ").strip()

    def _norm_nostrip_string(self, text: str):
        return ftfy.fix_text(text).replace("\xa0", " ")


class BasicNumberParser:
    parser = FSMParser(not_units=DEFAULT_NOT_UNITS.union(["st", "rd", "nd", "th"]))

    def parse(self, text: str, normed_text: str) -> Optional[ParsedNumber]:
        lst = self.parser.parse_value(normed_text)
        if len(lst) == 0:
            return None
        else:
            return lst[0]


class BasicDatetimeParser:
    def __init__(self) -> None:
        self.default_dt = datetime(MINYEAR, 1, 1)
        self.default_dt2 = datetime(MINYEAR + 3, 2, 28)

    def parse(self, text: str, normed_text: str) -> Optional[ParsedDatetimeRepr]:
        try:
            dt = dt_parse(normed_text, default=self.default_dt)
            year = dt.year
            month = dt.month
            day = dt.day

            if (
                dt.year == self.default_dt.year
                or dt.month == self.default_dt.month
                or dt.day == self.default_dt.day
            ):
                dt2 = dt_parse(text, default=self.default_dt2)

                if (
                    dt.year == self.default_dt.year
                    and dt2.year == self.default_dt2.year
                ):
                    year = None
                if (
                    dt.month == self.default_dt.month
                    and dt2.month == self.default_dt2.month
                ):
                    month = None
                if dt.day == self.default_dt.day and dt2.day == self.default_dt2.day:
                    day = None

            dt = ParsedDatetimeRepr(year=year, month=month, day=day)
        except (ParserError, TypeError, OverflowError):
            dt = None

        return dt



from __future__ import annotations

from operator import itemgetter
from typing import Callable, Mapping, TypeVar

from kgdata.models.ont_class import OntologyClass
from kgdata.models.ont_property import OntologyProperty

K = TypeVar("K")


def normalize_probs(
    target_values: list[tuple[K, float]],
    eps: float = 0.001,
    threshold: float = 0.0,
) -> list[tuple[K, float]]:
    """The predicted probabilities can be noisy due to numerical optimizations, i.e., equal variables
    can have slightly different scores. This function groups values that are close
    within the range [-eps, +eps] together, and replace them with the average value
    to reduce the noises.

    Args:
        target_values: probability of grounded values of a predicate
        eps: group values within [-eps, +eps] together
        threshold: remove values with probability less than threshold

    Return:
        A dictionary of normalized probabilities >= threshold
    """
    if eps == 0.0 and threshold == 0.0:
        return target_values

    norm_probs: dict[K, float] = {}
    lst = sorted([x for x in target_values if x[1] >= threshold], key=itemgetter(1))

    if len(lst) == 0:
        return []

    clusters = []
    pivot = 1
    clusters = [[lst[0]]]
    while pivot < len(lst):
        x = lst[pivot - 1][1]
        y = lst[pivot][1]
        if (y - x) <= eps:
            # same clusters
            clusters[-1].append(lst[pivot])
        else:
            # different clusters
            clusters.append([lst[pivot]])
        pivot += 1
    for cluster in clusters:
        avg_prob = sum([x[1] for x in cluster]) / len(cluster)
        for k, _prob in cluster:
            norm_probs[k] = avg_prob
    return sorted(norm_probs.items(), key=itemgetter(1), reverse=True)


def tiebreak(
    obj_and_score: list[tuple[K, float]],
    get_id: Callable[[K], str],
    id2popularity: Mapping[str, int],
    id2ent: Mapping[str, OntologyClass | OntologyProperty],
    eps: float = 1e-5,
):
    # decreasing sequence
    assert all(
        obj_and_score[i - 1][1] >= obj_and_score[i][1]
        for i in range(1, len(obj_and_score))
    ), "expect decreasing sequence"
    if len(obj_and_score) == 0:
        return

    best_score = obj_and_score[0][1]
    best_objs = []
    for obj, score in obj_and_score:
        if score < best_score:
            break
        best_objs.append(obj)

    best_idx = 0
    best_ent = id2ent[get_id(best_objs[best_idx])]

    for i in range(1, len(best_objs)):
        ent_id = get_id(best_objs[i])
        if (
            ent_id in best_ent.ancestors
            or id2popularity[ent_id] > id2popularity[best_ent.id]
        ):
            # the counting should be accumulated (the parent should have more count than the children) but
            # it is not, so we manually check it here
            best_idx = i
            best_ent = id2ent[ent_id]

    for i in range(len(best_objs)):
        if i != best_idx:
            # modified in place
            obj_and_score[i] = (obj_and_score[i][0], obj_and_score[i][1] - eps)



from dataclasses import dataclass, field

import gramsplus.core.literal_matchers as gcore_matcher


@dataclass
class LiteralMatcherConfig:
    """Configuration for literal matcher functions"""

    STRING: str = field(
        default=".string_exact_test", metadata={"help": "list of functions "}
    )
    QUANTITY: str = field(default=".quantity_test", metadata={"help": ""})
    GLOBECOORDINATE: str = field(default=".globecoordinate_test", metadata={"help": ""})
    TIME: str = field(default=".time_test", metadata={"help": ""})
    MONOLINGUAL_TEXT: str = field(
        default=".monolingual_exact_test", metadata={"help": ""}
    )
    ENTITY: str = field(default="", metadata={"help": ""})

    def to_rust(self) -> gcore_matcher.LiteralMatcherConfig:
        def py2ru(ident):
            if ident == "":
                return "always_fail_test"
            assert ident.startswith("."), ident
            return ident[1:]

        return gcore_matcher.LiteralMatcherConfig(
            py2ru(self.STRING),
            py2ru(self.QUANTITY),
            py2ru(self.GLOBECOORDINATE),
            py2ru(self.TIME),
            py2ru(self.MONOLINGUAL_TEXT),
            py2ru(self.ENTITY),
        )



from itertools import chain, combinations
from operator import itemgetter
from typing import Callable, Dict, List, Optional, Tuple

from graph.retworkx.api import digraph_all_simple_paths
from sm.dataset import FullTable

from gramsplus.semanticmodeling.postprocessing.cgraph import (
    CGEdge,
    CGEdgeTriple,
    CGNode,
    CGraph,
)


class PostProcessingSimplePath:
    def __init__(
        self,
        table: FullTable,
        cg: CGraph,
        edge_probs: Dict[CGEdgeTriple, float],
        threshold: float,
    ):
        self.table = table
        self.cg = cg
        self.edge_probs = edge_probs
        self.threshold = threshold

    def get_result(self) -> CGraph:
        """Select a simple path between nodes."""
        selected_edges = [e for e, p in self.edge_probs.items() if p >= self.threshold]
        subcg = self.cg.subgraph_from_edge_triples(selected_edges)
        subcg.remove_dangling_statement()
        if subcg.num_edges() == 0:
            # empty graph
            return subcg

        # ranking, prefer the shorter path, when there are multiple shorter path, select the one with the higher probs.
        return self.keep_one_simple_path_between_important_nodes(
            subcg, self.select_shorter_path, both_direction=True
        )

    def select_shorter_path(self, paths: List[List[CGEdgeTriple]]):
        """Prefer shorter path. When there are multiple shorter path, select the one with higher prob"""
        paths = sorted(paths, key=len)
        paths = [path for path in paths if len(path) == len(paths[0])]
        if len(paths) == 1:
            return paths[0]

        # multiple shorter paths
        path_probs = []
        for path in paths:
            path_prob = 0
            for uid, vid, eid in path:
                # edge = self.cg.get_edge_between_nodes(uid, vid, eid)
                path_prob += self.edge_probs[uid, vid, eid]
            path_probs.append(path_prob)
        path, path_prob = max(zip(paths, path_probs), key=itemgetter(1))
        return path

    def keep_one_simple_path_between_important_nodes(
        self,
        cg: CGraph,
        ranking_fn: Optional[
            Callable[[List[List[CGEdgeTriple]]], List[CGEdgeTriple]]
        ] = None,
        both_direction: bool = True,
    ):
        """
        Let important nodes be columns and context values. If there is more than one path between the important nodes,
        we only keep one. Note that this function doesn't remove a literal value if it is value of a property of a statement
        that appear in the chosen path.

        The default ranking function is select the path with shorter length. Supply your own function for better result.
        The both direction function will select the correct one
        """
        # figure out the important nodes
        important_nodes = set()
        for u in cg.nodes():
            if isinstance(u, CGColumnNode) or (
                isinstance(u, (CGEntityValueNode, CGLiteralValueNode))
                and u.is_in_context
            ):
                important_nodes.add(u.id)

        if ranking_fn is None:
            ranking_fn = lambda paths: sorted(paths, key=lambda path: len(path))[0]

        # select one path if there are multiple edges
        selected_edges = set()
        for uid, vid in combinations(important_nodes, 2):
            forward_paths = [
                [(e.source, e.target, e.predicate) for e in path]
                for path in digraph_all_simple_paths(cg, uid, vid, cutoff=6)
                if not any(edge.source in important_nodes for edge in path[1:])
            ]
            backward_paths = [
                [(e.source, e.target, e.predicate) for e in path]
                for path in digraph_all_simple_paths(cg, vid, uid, cutoff=6)
                if not any(edge.source in important_nodes for edge in path[1:])
            ]
            if both_direction:
                if len(forward_paths) + len(backward_paths) > 0:
                    selected_edges = selected_edges.union(
                        ranking_fn(forward_paths + backward_paths)
                    )
            else:
                if len(forward_paths) > 0:
                    selected_edges = selected_edges.union(ranking_fn(forward_paths))
                if len(backward_paths) > 0:
                    selected_edges = selected_edges.union(ranking_fn(backward_paths))

        # new_sg = SemanticGraphConstructor.get_sg_subgraph(sg, selected_edges)
        new_cg = cg.subgraph_from_edge_triples(selected_edges)

        # add missing statement values
        for stmt in new_cg.nodes():
            if not isinstance(stmt, CGStatementNode):
                continue

            # ((uid, _, eid),) = new_cg.in_edges(stmt.id)
            (us_edge,) = new_cg.in_edges(stmt.id)
            if not any(
                sv_edge.predicate == us_edge.predicate
                for sv_edge in new_cg.out_edges(stmt.id)
            ):
                # we do not have the statement property in the graph. get it from the original graph
                (sv_edge,) = [
                    e for e in cg.out_edges(stmt.id) if e.predicate == us_edge.predicate
                ]
                if not new_cg.has_node(sv_edge.target):
                    new_cg.add_node(cg.get_node(sv_edge.target).clone())
                new_cg.add_edge(sv_edge.clone())

        return new_cg



from __future__ import annotations

from collections import defaultdict
from copy import deepcopy
from dataclasses import dataclass, field
from pathlib import Path
from typing import Literal, Optional

import ray
from libraries.gp.gp.actors.data._db import KGDB, DBActor, DBActorArgs
from gramsplus.distantsupervision.make_dataset.prelude import (
    CombinedFilter,
    CombinedFilterArgs,
    EntityRecognitionV1,
    EntityRecognitionV1Args,
    FilterByEntType,
    FilterByEntTypeArgs,
    FilterByHeaderColType,
    FilterByHeaderColTypeArgs,
    FilterFn,
    FilterNotEntCol,
    FilterRegex,
    FilterRegexArgs,
    FilterV1,
    FilterV1Args,
    LabelFn,
    LabelV1,
    LabelV1Args,
    LabelV2,
    LabelV2Args,
    NoFilter,
    NoTransform,
    TransformFn,
    TransformV1,
    TransformV1Args,
    TransformV2,
)
from gramsplus.entity_linking.candidate_recognition.heuristic_model import (
    HeuristicCanRegArgs,
)
from gramsplus.semanticmodeling.text_parser import TextParser
from ream.cache_helper import MemBackend
from ream.dataset_helper import DatasetList, DatasetQuery
from ream.params_helper import NoParams
from ream.prelude import BaseActor, Cache
from ream.workspace import ReamWorkspace
from sm.dataset import Dataset, Example, FullTable
from sm.inputs.prelude import EntityIdWithScore
from sm.misc.ray_helper import get_instance, ray_map, ray_put
from sm.namespaces.utils import KGName
from sm_datasets.datasets import Datasets


@dataclass
class AutoLabelDataActorArgs:
    dataset_dir: Path = field(
        metadata={
            "help": "Path to the directory containing datasets of linked tables",
        }
    )
    skip_non_unique_mention: bool = field(
        default=False,
        metadata={
            "help": "Skip tables with non-unique mention",
        },
    )
    skip_column_with_no_type: bool = field(
        default=False,
        metadata={
            "help": "Column that auto-labeler cannot find any entity type will be skipped"
        },
    )
    recog_method: Literal["recog_v1"] = field(
        default="recog_v1",
        metadata={
            "help": "Entity recognition method to use",
            "variants": {"recog_v1": EntityRecognitionV1},
        },
    )
    recog_v1: EntityRecognitionV1Args = field(
        default_factory=EntityRecognitionV1Args,
        metadata={"help": "Entity recognition v1 arguments"},
    )
    filter_method: Literal[
        "no_filter",
        "filter_non_ent_col",
        "filter_regex",
        "filter_v1",
        "filter_ent_type",
        "filter_header_col_type",
        "filter_combined",
    ] = field(
        default="filter_regex",
        metadata={
            "help": "Filter method to use",
            "variants": {
                "no_filter": NoFilter,
                "filter_non_ent_col": FilterNotEntCol,
                "filter_regex": FilterRegex,
                "filter_v1": FilterV1,
                "filter_ent_type": FilterByEntType,
                "filter_header_col_type": FilterByHeaderColType,
                "filter_combined": CombinedFilter,
            },
        },
    )
    filter_regex: FilterRegexArgs = field(
        default_factory=FilterRegexArgs,
        metadata={"help": "Filter regex arguments"},
    )
    filter_ent_type: FilterByEntTypeArgs = field(
        default_factory=FilterByEntTypeArgs,
        metadata={"help": "Filter by entity type arguments"},
    )
    filter_not_ent_col: NoParams = field(
        default_factory=NoParams,
        metadata={"help": "Filter not entity column arguments"},
    )
    filter_header_col_type: Optional[FilterByHeaderColTypeArgs] = field(
        default=None,
        metadata={"help": "Filter by header col type arguments"},
    )
    filter_combined: CombinedFilterArgs = field(
        default_factory=CombinedFilterArgs,
        metadata={"help": "Combined filter arguments"},
    )
    filter_v1: FilterV1Args = field(
        default_factory=FilterV1Args,
        metadata={"help": "Filter v1 arguments"},
    )
    transform_method: Literal["transform_v1", "transform_v2", "no_transform"] = field(
        default="transform_v1",
        metadata={
            "help": "Transformation method to use",
            "variants": {
                "transform_v1": TransformV1,
                "no_transform": NoTransform,
                "transform_v2": TransformV2,
            },
        },
    )
    transform_v1: Optional[TransformV1Args] = field(
        default=None,
        metadata={"help": "Transformation v1 arguments"},
    )
    transform_v2: Optional[TransformV1Args] = field(
        default=None,
        metadata={"help": "Transformation v2 arguments"},
    )
    label_method: Literal["label_v1", "label_v2"] = field(
        default="label_v1",
        metadata={
            "help": "Label method to use",
            "variants": {"label_v1": LabelV1, "label_v2": LabelV2},
        },
    )
    label_v1: Optional[LabelV1Args] = None
    label_v2: Optional[LabelV2Args] = None


@dataclass
class AutoLabeledTable:
    table: FullTable
    entity_columns: list[int]
    entity_column_types: list[list[EntityIdWithScore]]

    def to_dict(self):
        return {
            "table": self.table.to_dict(),
            "entity_columns": self.entity_columns,
            "entity_column_types": [
                [e.to_dict() for e in coltypes] for coltypes in self.entity_column_types
            ],
        }

    @classmethod
    def from_dict(cls, obj: dict):
        return cls(
            table=FullTable.from_dict(obj["table"]),
            entity_columns=obj["entity_columns"],
            entity_column_types=[
                [EntityIdWithScore.from_dict(e) for e in coltypes]
                for coltypes in obj["entity_column_types"]
            ],
        )


class AutoLabeledDataActor(BaseActor[AutoLabelDataActorArgs]):
    VERSION = 116

    def __init__(self, params: AutoLabelDataActorArgs, db_actor: DBActor):
        super().__init__(params, [db_actor])
        self.db_actor = db_actor

    @Cache.cache(backend=MemBackend())
    def __call__(self, dsquery: str) -> DatasetList[AutoLabeledTable]:
        parsed_dsquery = DatasetQuery.from_string(dsquery)
        tables = self.process_dataset(parsed_dsquery.dataset)
        return parsed_dsquery.select_list(tables)

    def is_autolabel_dataset(self, dsquery: str) -> bool:
        parsed_dsquery = DatasetQuery.from_string(dsquery)
        return (self.params.dataset_dir / parsed_dsquery.dataset).exists()

    @Cache.cache(
        backend=Cache.cls.dir(
            cls=DatasetList,
            dirname="process_{dataset}",
            compression="lz4",
            mem_persist=True,
            log_serde_time=True,
        ),
    )
    def process_dataset(self, dataset: str) -> DatasetList[AutoLabeledTable]:
        examples = Dataset(self.params.dataset_dir / dataset).load()

        # when we change the databases, sometimes the entities are not the same.
        kgdb = self.get_kgdb(dataset)
        entity_labels = kgdb.pydb.entity_labels.cache()
        props = kgdb.pydb.props.cache()
        redirections = kgdb.pydb.entity_redirections.cache()
        examples = Datasets().fix_redirection(
            examples,
            entity_labels,
            props,
            redirections,
            kgdb.kgns,
        )

        if len(examples) > 1000:
            selfargs = ray_put(
                (self.get_working_fs().root, self.params, self.db_actor.params)
            )

            def ray_process_example(
                selfargs, dataset: str, ream_args: dict, ex: Example[FullTable]
            ):
                ReamWorkspace.init_from_dict(ream_args)
                self = AutoLabeledDataActor.get_instance(selfargs)
                return self.process_example(
                    ex,
                    self.get_er(),
                    self.get_filter(dataset),
                    self.get_transformation(dataset),
                    self.get_label(dataset),
                    TextParser.default(),
                )

            ream_args_ref = ray.put(ReamWorkspace.get_instance().to_dict())
            output = ray_map(
                ray_process_example,
                [(selfargs, dataset, ream_args_ref, ex) for ex in examples],
                verbose=True,
                desc="generate auto-label dataset",
                auto_shutdown=True,
                is_func_remote=False,
            )
            autolabel_tables = [x for x in output if x is not None]
        else:
            er = self.get_er()
            fil = self.get_filter(dataset)
            map = self.get_transformation(dataset)
            labeler = self.get_label(dataset)

            text_parser = TextParser.default()
            autolabel_tables = []
            for ex in examples:
                newtable = self.process_example(ex, er, fil, map, labeler, text_parser)
                if newtable is not None:
                    autolabel_tables.append(newtable)
        return DatasetList(dataset, autolabel_tables)

    def process_example(
        self,
        ex: Example[FullTable],
        er: EntityRecognitionV1,
        fil: FilterFn,
        map: TransformFn,
        labeler: LabelFn,
        text_parser: TextParser,
    ) -> Optional[AutoLabeledTable]:
        table = ex.table
        if self.params.skip_non_unique_mention and has_non_unique_mention(table):
            return None

        # recognize entity columns
        entity_columns = er.recognize(table)
        entity_columns = fil.filter(table, entity_columns)

        table = map.transform(table, entity_columns)
        entity_column_types = labeler.label(table, entity_columns)

        if self.params.skip_column_with_no_type:
            valid_cols = [
                i for i in range(len(entity_columns)) if len(entity_column_types[i]) > 0
            ]
            entity_columns = [entity_columns[i] for i in valid_cols]
            entity_column_types = [entity_column_types[i] for i in valid_cols]

        if len(entity_columns) == 0:
            return None

        newtable = normalize_table(ex.table, text_parser)
        return AutoLabeledTable(newtable, entity_columns, entity_column_types)

    @Cache.cache(backend=MemBackend())
    def get_er(self):
        if self.params.recog_method == "recog_v1":
            return EntityRecognitionV1(self.params.recog_v1)
        raise NotImplementedError()

    @Cache.cache(backend=MemBackend())
    def get_filter(self, dataset: str):
        logfile = self.get_working_fs().root / f"logs/{dataset}/filter.log"
        kgdb = self.get_kgdb(dataset)

        if self.params.filter_method == "filter_regex":
            return FilterRegex(self.params.filter_regex, logfile)
        if self.params.filter_method == "filter_v1":
            return FilterV1(
                self.params.filter_v1,
                kgdb.pydb.entities.cache(),
                logfile,
            )
        if self.params.filter_method == "filter_combined":
            filters = [
                FilterRegex(self.params.filter_combined.regex, logfile),
                FilterByEntType(
                    self.params.filter_combined.ignore_types,
                    kgdb.pydb.entities.cache(),
                    kgdb.pydb.classes.cache(),
                    logfile,
                ),
            ]
            if self.params.filter_combined.header_col_type is not None:
                filters.append(
                    FilterByHeaderColType(
                        self.params.filter_combined.header_col_type,
                        self.get_label(dataset),
                        logfile,
                    )
                )
            return CombinedFilter(
                filters,
                logfile,
            )
        if self.params.filter_method == "no_filter":
            return NoFilter()
        if self.params.filter_method == "filter_non_ent_col":
            return FilterNotEntCol(HeuristicCanRegArgs(), logfile)
        raise NotImplementedError()

    @Cache.cache(backend=MemBackend())
    def get_transformation(self, dataset: str):
        logfile = self.get_working_fs().root / f"logs/{dataset}/transform.log"
        kgdb = self.get_kgdb(dataset)
        if self.params.transform_method == "transform_v1":
            assert self.params.transform_v1 is not None
            return TransformV1(
                self.params.transform_v1,
                kgdb.pydb.entities.cache(),
                kgdb.pydb.classes.cache(),
                logfile,
            )
        if self.params.transform_method == "transform_v2":
            assert self.params.transform_v2 is not None
            return TransformV2(
                self.params.transform_v2,
                kgdb.pydb.entities.cache(),
                kgdb.pydb.classes.cache(),
                logfile,
            )

        if self.params.transform_method == "no_transform":
            return NoTransform()
        raise NotImplementedError()

    @Cache.cache(backend=MemBackend())
    def get_label(self, dataset: str) -> LabelFn:
        kgdb = self.get_kgdb(dataset)
        if self.params.label_method == "label_v1":
            assert self.params.label_v1 is not None
            return LabelV1(
                self.params.label_v1,
                kgdb.pydb.entities.cache(),
                kgdb.pydb.entity_pagerank.cache(),
                kgdb.pydb.classes.cache(),
            )
        if self.params.label_method == "label_v2":
            assert self.params.label_v2 is not None
            return LabelV2(
                self.params.label_v2,
                kgdb.pydb.entities.cache(),
                kgdb.pydb.entity_pagerank.cache(),
                kgdb.pydb.classes.cache(),
            )
        raise NotImplementedError()

    @Cache.cache(backend=MemBackend())
    def get_kgname(self, dataset: str):
        if dataset.startswith("wt"):
            return KGName.Wikidata
        raise NotImplementedError(dataset)

    def get_kgdb(self, dataset: str) -> KGDB:
        kgname = self.get_kgname(dataset)
        return self.db_actor.kgdbs[kgname]

    @staticmethod
    def get_instance(
        args: AutoLabeledDataActor | tuple[str, AutoLabelDataActorArgs, DBActorArgs]
    ):
        if isinstance(args, AutoLabeledDataActor):
            return args
        return get_instance(
            lambda: AutoLabeledDataActor(args[1], DBActor(args[2])),
            f"autolabeled_actor:{args[0]}",
        )


def has_non_unique_mention(table: FullTable) -> bool:
    """Check if the example table has the same mention at different cells linked to different entities"""
    col2mention = defaultdict(lambda: defaultdict(set))

    for ri, ci, links in table.links.enumerate_flat_iter():
        if len(links) == 0:
            continue

        text = table.table[ri, ci]
        assert isinstance(text, str), text

        for link in links:
            mention = text[link.start : link.end]
            if len(mention) > 0:
                col2mention[ci][mention].update(link.entities)
                if len(col2mention[ci][mention]) > 1:
                    return True

    return False


def normalize_table(oldtable: FullTable, text_parser: TextParser) -> FullTable:
    table = deepcopy(oldtable)
    for col in table.table.columns:
        assert col.name is not None
        col.name = text_parser._norm_string(col.name)

    # normalize cells and links
    for ci, col in enumerate(table.table.columns):
        for ri, cell in enumerate(col.values):
            if isinstance(cell, str):
                newcell = text_parser._norm_string(cell)
                col.values[ri] = newcell

                if newcell != cell:
                    # adjust the links
                    for link in table.links[ri, ci]:
                        if (
                            cell[link.start : link.end]
                            != newcell[link.start : link.end]
                        ):
                            # the new mention is different from the old mention
                            before = text_parser._norm_nostrip_string(
                                cell[: link.start]
                            ).lstrip()
                            mention = text_parser._norm_nostrip_string(
                                cell[link.start : link.end]
                            )
                            if len(before) == 0 and mention.lstrip() != mention:
                                mention = mention.lstrip()
                            after = text_parser._norm_nostrip_string(
                                cell[link.end :]
                            ).rstrip()
                            if len(after) == 0 and mention.rstrip() != mention:
                                mention = mention.rstrip()
                            if before + mention + after != newcell:
                                raise NotImplementedError(
                                    f"Haven't implemented fixing where part of the mention has been changed. Recovered string: `{before+mention+after}` - transformed string: `{newcell}`"
                                )
                            link.start = len(before)
                            link.end = len(before) + len(mention)
    return table



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Callable, Optional

import numpy as np
import pandas as pd
import torch
from ream.data_model_helper import (
    EncodedSingleMasked2DNumpyArray,
    EncodedSingleNumpyArray,
)
from sm.misc.funcs import get_decoder
from torch.utils.data import Dataset
from torch.utils.data._utils.collate import collate, default_collate_fn_map


class ColumnarDataset(Dataset):
    """A columnar dataset"""

    def __init__(
        self,
        columns: dict[str, np.ndarray | list | Feat],
        dtypes: Optional[dict[str, Any]] = None,
        collate_fn: Optional[Callable] = None,
        references: Optional[dict] = None,
        name: str = "",
    ):
        self.name = name
        self.collate_fn = collate_fn
        self.columns = columns
        self.size = len(next(iter(self.columns.values())))
        self.dtypes = dtypes
        self.references = references or {}

        if dtypes is not None:
            for name, feat in self.columns.items():
                if name in dtypes and isinstance(feat, np.ndarray):
                    self.columns[name] = feat.astype(dtypes[name])
                    dtypes.pop(name)

    def update_size(self):
        self.size = len(next(iter(self.columns.values())))

    def __len__(self):
        return self.size

    def __getitem__(self, idx: int | slice):
        obj = {name: feat[idx] for name, feat in self.columns.items()}
        if isinstance(idx, slice):
            return ColumnarDataset(
                obj,
                dtypes=self.dtypes,
                collate_fn=self.collate_fn,
                references=self.references,
                name=self.name,
            )
        return obj

    def to_df(self):
        cols = {}
        for name, feat in self.columns.items():
            if not isinstance(feat, np.ndarray):
                raise ValueError(
                    f"Column {name} is not a numpy array, cannot convert dataset to dataframe"
                )
            if len(feat.shape) > 2:
                raise ValueError(
                    f"Column {name} has more than 2 dimensions, cannot convert dataset to dataframe"
                )
            if len(feat.shape) == 2:
                for i in range(feat.shape[1]):
                    cols[f"{name}_{i}"] = feat[:, i]
            else:
                cols[name] = feat
        return pd.DataFrame(cols)


@dataclass
class Feat:
    def __getitem__(self, idx: int | slice):
        raise NotImplementedError()

    def __len__(self):
        raise NotImplementedError()


@dataclass
class EmbeddingFeat(Feat):
    """For nd-array, this code works for 3d array as well.

    Example:
        >>> a = np.arange(16).reshape((4,4))
        >>> idx = [[0, 1, 3], [1, 3, 2]]
        >>> a[idx]

    """

    value: np.ndarray | list
    embs: np.ndarray
    decoder: Optional[list[str]] = None

    def __getitem__(self, idx: int | slice):
        if not isinstance(idx, int):
            return EmbeddingFeat(self.value[idx], self.embs, self.decoder)
        return self.embs[self.value[idx]]

    def __len__(self):
        return len(self.value)

    @staticmethod
    def from_encoded_single_numpy_array(
        array: EncodedSingleNumpyArray | EncodedSingleMasked2DNumpyArray,
        emb_fn: Callable[[list[str]], np.ndarray],
    ):
        return EmbeddingFeat(array.value, emb_fn(array.decoder), array.decoder)

    def to_array(self):
        return self.embs[self.value]

    def merge(self, other: EmbeddingFeat):
        # haven't handle non-decoder yet
        assert self.decoder is not None and other.decoder is not None
        self_encoder = {v: i for i, v in enumerate(self.decoder)}

        other_valuemap = [10000000 for _ in other.decoder]
        new_embs = []

        for i, v in enumerate(other.decoder):
            if v not in self_encoder:
                self_encoder[v] = len(self_encoder)
                new_embs.append(other.embs[i])
            other_valuemap[i] = self_encoder[v]

        value = np.concatenate(
            [self.value, [other_valuemap[v] for v in other.value]], axis=0
        )
        if len(new_embs) > 0:
            embs = np.concatenate([self.embs, np.stack(new_embs, axis=0)], axis=0)
        else:
            embs = self.embs
        return EmbeddingFeat(value, embs, get_decoder(self_encoder))


@dataclass
class DynSize:
    value: Any

    @staticmethod
    def collate_fn(batch, *, collate_fn_map=None):
        return torch.stack([torch.as_tensor(b.value) for b in batch])


@dataclass
class SmallDynSize:
    value: Any

    @staticmethod
    def collate_fn(batch, *, collate_fn_map=None):
        size = max(len(b.value) for b in batch)
        return [
            torch.as_tensor(np.pad(b.value, [(0, size - b.value.shape[0]), (0, 0)]))
            for b in batch
        ]


collate_fn_map = default_collate_fn_map.copy()
collate_fn_map[DynSize] = DynSize.collate_fn
collate_fn_map[SmallDynSize] = SmallDynSize.collate_fn


def extended_collate_fn(batch):
    return collate(batch, collate_fn_map=collate_fn_map)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Literal, Optional, Sequence

from gp.actors.data.prelude import KGDB, DataActor, GPExample, KGDBArgs
from gp.actors.sm.can_graph import CanGraphActor
from gp.entity_linking.cangen.common import TableCanGenResult
from gp.misc.appconfig import AppConfig
from gp.misc.evaluation.el_osin_mixin import EntityLinkingOsinActorMixin
from gp.misc.evaluation.sm_osin_mixin import SMOsinActorMixin
from gp.misc.evaluation.sm_wikidata import SemanticModelHelper
from gp.semanticmodeling.models.pretrained_models import (
    PretrainedCPAModel,
    PretrainedCTAModel,
)
from gp.semanticmodeling.postprocessing.cgraph import CGraph
from gp.semanticmodeling.postprocessing.greedy_known_targets import GreedyKnownTargetsFn
from gp.semanticmodeling.postprocessing.interface import EdgeProb, NodeProb
from gp.semanticmodeling.postprocessing.steiner_tree import SteinerTreeFn
from gramsplus.actors.data import PredictionTargets
from gramsplus.core.algorithms import CanGraphExtractedResult
from ream.cache_helper import Cache, FileBackend, MemBackend
from ream.prelude import BaseActor
from sm.dataset import Example, FullTable
from sm.misc.ray_helper import get_instance, ray_map, ray_put
from sm.namespaces.wikidata import WikidataNamespace
from sm.outputs.semantic_model import SemanticModel
from sm.prelude import O

PostprocessingMethod = Literal["greedy_known_targets", "steiner_tree"]
OnUntypeSourceColumnNode = Literal["create-class", "remove-link"]


@dataclass
class GraphInferenceActorArgs:
    postprocessing: PostprocessingMethod = field(
        default="steiner_tree",
        metadata={
            "help": "The post-processing method to use",
        },
    )
    on_untype_source_column_node: OnUntypeSourceColumnNode = field(
        default="create-class",
        metadata={
            "help": "How to handle untyped source node",
        },
    )


class GraphInferenceActor(
    SMOsinActorMixin[GraphInferenceActorArgs],
    EntityLinkingOsinActorMixin[GraphInferenceActorArgs],
    BaseActor[GraphInferenceActorArgs],
):
    """Graph Inference Actor"""

    VERSION = 101
    EXP_NAME = "Graph Inference"
    EXP_VERSION = 100

    def __init__(
        self,
        params: GraphInferenceActorArgs,
        cangraph_actor: CanGraphActor,
        pretrained_cpa_model: PretrainedCPAModel,
        pretrained_cta_model: PretrainedCTAModel,
    ):
        super().__init__(
            params,
            dep_actors=[
                cangraph_actor,
                pretrained_cpa_model,
                pretrained_cta_model,
            ],
        )

        self.pretrained_cpa_model = pretrained_cpa_model
        self.pretrained_cta_model = pretrained_cta_model

        self.cangraph_actor = cangraph_actor
        self.canrank_actor = cangraph_actor.canrank_actor
        self.db_actor = self.canrank_actor.db_actor

    # @Cache.cache(
    #     backend=MemBackend(),
    #     # backend=Cache.file.pickle(mem_persist=True),
    #     cache_ser_args={
    #         "cta_threshold": lambda x: str(x),
    #         "cpa_threshold": lambda x: str(x),
    #     },
    #     cache_args=["dataset", "cpa_threshold", "cta_threshold"],
    # )
    def batch_call(
        self,
        exs: list[GPExample],
        cpa_threshold: float = 0.5,
        cta_threshold: float = 0.5,
        known_targets: Optional[list[PredictionTargets]] = None,
    ):
        ex_cangraphs = self.cangraph_actor.batch_call(exs)
        edge_probs = self.pretrained_cpa_model.batch_call(exs)
        node_probs = self.pretrained_cta_model.batch_call(exs)

        if self.params.postprocessing == "greedy_known_targets":
            assert known_targets is not None
            targets = known_targets
            assert len(targets) == len(exs)
            postprocessing_args = []
            for target in targets:
                postprocessing_args.append(
                    {
                        "edge_targets": target.cpa,
                        "type_targets": target.cta,
                        "cpa_threshold": cpa_threshold,
                        "cta_threshold": cta_threshold,
                    }
                )
        else:
            postprocessing_args = [
                {
                    "cpa_threshold": cpa_threshold,
                    "cta_threshold": cta_threshold,
                    "cta_score_offset": self.pretrained_cta_model.cta_score_offset(),
                }
                for _ in range(len(exs))
            ]

        pred_sms = self.predict_sm(
            self.db_actor.kgdbs[exs[0].kgname],
            exs,
            ex_cangraphs,
            edge_probs,
            node_probs,
            self.params.postprocessing,
            postprocessing_args,
            self.params.on_untype_source_column_node,
        )
        return pred_sms

    def get_semantic_models(
        self, exs: list[GPExample], **kwargs
    ) -> list[SemanticModel]:
        return self.batch_call(exs, **kwargs)

    def get_candidate_entities(self, exs: list[GPExample]) -> list[TableCanGenResult]:
        return self.cangraph_actor.get_candidate_entities(exs)

    def get_edge_probs(self, exs: list[GPExample], **kwargs) -> list[EdgeProb]:
        db = self.db_actor.kgdbs[exs[0].kgname]
        props = db.pydb.props.cache()

        out: list[EdgeProb] = []
        ex_cangraphs = self.cangraph_actor.batch_call(exs)
        for cangraph, edge_probs in zip(
            ex_cangraphs, self.pretrained_cpa_model.batch_call(exs, True)
        ):
            new_edge_probs = {}
            for (source, target, edge), eprob in edge_probs.items():
                source = cangraph.nodes[int(source)]
                target = cangraph.nodes[int(target)]

                if source.is_column():
                    scol = source.try_as_column()
                    assert scol is not None
                    slbl = f"{scol.label} ({scol.column})"
                elif source.is_entity():
                    sent = source.try_as_entity()
                    assert sent is not None
                    slbl = f"ent:{sent.entity_id}"
                elif source.is_literal():
                    slit = source.try_as_literal()
                    assert slit is not None
                    slbl = f"literal:{slit.value.to_string_repr()}"
                else:
                    assert source.is_statement()
                    sstm = source.try_as_statement()
                    assert sstm is not None
                    slbl = f"statement:{sstm.id}"

                if target.is_column():
                    scol = target.try_as_column()
                    assert scol is not None
                    tlbl = f"{scol.label} ({scol.column})"
                elif target.is_entity():
                    sent = target.try_as_entity()
                    assert sent is not None
                    tlbl = f"ent:{sent.entity_id}"
                elif target.is_literal():
                    slit = target.try_as_literal()
                    assert slit is not None
                    tlbl = f"literal:{slit.value.to_string_repr()}"
                else:
                    assert target.is_statement()
                    sstm = target.try_as_statement()
                    assert sstm is not None
                    tlbl = f"statement:{sstm.id}"

                new_edge_probs[(slbl, tlbl, str(props[edge]))] = eprob
            out.append(new_edge_probs)

        return out

    def get_node_probs(self, exs: list[GPExample], **kwargs):
        return self.pretrained_cta_model.batch_call(exs, True)

    def predict_sm(
        self,
        kgdb: KGDB,
        examples: list[GPExample],
        cangraphs: list[CanGraphExtractedResult],
        edge_probs: list[EdgeProb],
        node_probs: list[NodeProb],
        postprocessing_method: PostprocessingMethod = "steiner_tree",
        postprocessing_args: Optional[Sequence[dict]] = None,
        on_untype_source_column_node: OnUntypeSourceColumnNode = "create-class",
    ):
        using_ray = len(examples) > 1 and not AppConfig.get_instance().is_ray_disable
        dbargs = ray_put(kgdb.args, using_ray)
        using_ray = False
        verbose = False

        return ray_map(
            postprocessing,
            [
                (
                    dbargs,
                    ex.table,
                    cangraphs[ei],
                    edge_probs[ei],
                    node_probs[ei],
                    postprocessing_method,
                    postprocessing_args[ei] if postprocessing_args is not None else {},
                    on_untype_source_column_node,
                )
                for ei, ex in enumerate(examples)
            ],
            desc="predict sm postprocessing",
            verbose=verbose,
            using_ray=using_ray,
            is_func_remote=False,
        )


def postprocessing(
    kgdb: KGDB | KGDBArgs,
    table: FullTable,
    cangraph: CanGraphExtractedResult,
    edge_probs: EdgeProb,
    node_probs: NodeProb,
    postprocessing_method: PostprocessingMethod,
    postprocessing_extraargs: dict,
    on_untype_source_column_node: OnUntypeSourceColumnNode,
):
    db = KGDB.get_instance(kgdb).pydb
    sm_helper = get_instance(
        lambda: SemanticModelHelper(
            db.entity_labels,
            db.props,
            WikidataNamespace.create(),
        ),
        __name__ + ".sm_helper",
    )

    if postprocessing_method == "steiner_tree":
        pp = SteinerTreeFn(
            table, CGraph.from_rust(cangraph), **postprocessing_extraargs
        )
    elif postprocessing_method == "greedy_known_targets":
        pp = GreedyKnownTargetsFn(
            CGraph.from_rust(cangraph), **postprocessing_extraargs
        )
    else:
        raise NotImplementedError(postprocessing_method)

    ppres = pp(node_probs, edge_probs)

    # convert cgraph into nodes/edges
    nodes = {}
    cpa = [(int(source), int(target), key) for source, target, key in ppres.edges]
    cta = ppres.nodes

    # gather nodes that will be in the semantic model
    uids = set()
    for source, target, prop_id in cpa:
        uids.add(source)
        uids.add(target)

    for u in cangraph.nodes:
        if u.is_column() and u.id not in uids:
            u = u.try_as_column()
            assert u is not None
            if u.column in cta:
                uids.add(u.id)

    for uid in uids:
        assert uid not in nodes
        u = cangraph.nodes[uid]
        if u.is_column():
            u = u.try_as_column()
            assert u is not None
            uprime = O.DataNode(
                col_index=u.column,
                label=str(u.label),
            )
        elif u.is_entity():
            u = u.try_as_entity()
            assert u is not None
            uprime = O.LiteralNode(
                value=sm_helper.kgns.id_to_uri(u.entity_id),
                readable_label=sm_helper.get_ent_label(u.entity_id),
                is_in_context=u.is_in_context,
                datatype=O.LiteralNodeDataType.Entity,
            )
        elif u.is_literal():
            u = u.try_as_literal()
            assert u is not None
            # TODO: how to ensure to_string_repr for the same value is the same?
            uprime = O.LiteralNode(
                value=u.value.to_string_repr(),
                readable_label=None,
                is_in_context=u.is_in_context,
                datatype=O.LiteralNodeDataType.String,
            )
        else:
            assert u.is_statement()
            u = u.try_as_statement()
            assert u is not None
            uprime = O.ClassNode(
                abs_uri=sm_helper.kgns.statement_uri,
                rel_uri=sm_helper.kgns.get_rel_uri(sm_helper.kgns.statement_uri),
                approximation=False,
                readable_label="wikibase:Statement",
            )
        nodes[uid] = uprime

    return sm_helper.create_sm(
        nodes, cpa, cta, on_untype_source_column_node=on_untype_source_column_node
    )
